#include <assert.h>
#include <ctype.h>
#include <malloc.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>

#include "Formula.h"

/*#define DEBUG*/

static char SYNTAX_ERROR[] = "Syntax error";
static char NO_CORRESPONDED_PARENTHESIS_END[] = "No corresponded parenthesis found";
static char BOOL_OP_ERROR[] = "Boolean operation may not have REAL or DOUBLE operands";
static char ONLY_VAR_IN_FUNC[] = "Only variables may be passed as arguments to MAX and MIN functions";
static char DIVIDE_BY_ZERO[] = "Divide by zero";

/*************************  FREE Formula functions **************/
static void FreeIndex(Index *index)
{
	Index *pred;
	while(index)
	{
		pred = index;
		index = index->next;
		free(pred);
	}
}

Value* MakeValues(int amount)
{
	int i;
	Value *value;

	if((value = malloc(sizeof(Value) * amount)) == NULL)
		return NULL;
	for(i = 0; i < amount; i++)
		value[i].indexes = NULL;
	return value;
}

Value* MakeValue(void) { return MakeValues(1);}

ValuesProcArray *MakeValuesProcArray(int nproc, int amount_in_each)
{
	ValuesProcArray *values;
	int i;

	if((values = malloc(sizeof(ValuesProcArray))) == NULL)
		return NULL;
	values->proc_amount = nproc;
	if((values->array = calloc(sizeof(ValuesArray), nproc)) == NULL)
	{
		free(values);
		return NULL;
	}
	if(amount_in_each)
		for(i = 0; i < nproc; i++)
		{
			if((values->array[i].values = MakeValues(amount_in_each)) == NULL)
			{
				FreeValuesProcArray(values);
				return NULL;
			}
			values->array[i].amount = amount_in_each;
		}
	return values;
}

static void CopyRIValue(Value *destination, Value *source)
{
	if(destination == source)
		return;
	if(destination->indexes)
	{
		free(destination->indexes);
		destination->indexes = NULL;
	}
	memcpy(destination, source, sizeof(Value));
	source->indexes = NULL;
}

void FreeValues(Value *value, int amount)
{
#ifdef DEBUG
	fprintf(stderr, "FreeValues, amount = %d\n", amount);
#endif
	if((value == NULL) || (amount == 0))
		return;

	while(amount-- > 0)
	{
		if(value[amount].indexes)
		{
#ifdef DEBUG
			fprintf(stderr, "FreeValues: value[%d] has indexes\n", amount);
#endif
			free(value[amount].indexes);
		}
		if(value[amount].type == CHARACTER_T)
			free(value[amount].value.char_value);
	}
	free(value);
}

void FreeValue(Value *value)
{
	FreeValues(value, 1);
}

void FreeValuesProcArray(ValuesProcArray *values)
{
	if(values)
	{
		int i;
		for(i = 0; i < values->proc_amount; i++)
			FreeValues(values->array[i].values, values->array[i].amount);
		free(values);
	}
}

static void FreeVariable(Variable *variable)
{
	FreeIndex(variable->index);
	free(variable);
}

static void FreeOperator(Operator *noperator)
{
	FreeFormula(noperator->left);
	FreeFormula(noperator->right);
	free(noperator);
}

static void FreeFunction(Function *function)
{
	FreeFormula(function->operand);
	free(function);
}

void FreeFormula(Formula *formula)
{
	if(formula == NULL)
		return;

	switch(formula->type)
	{
	case NoItem:
		break;
	case VariableItem:
		FreeVariable(formula->item.variable);
		break;
	case OperatorItem:
		FreeOperator(formula->item.noperator);
		break;
	case ValueItem:
		FreeValue(formula->item.value);
		break;
	case FunctionItem:
		FreeFunction(formula->item.function);
		break;
	}
	free(formula);
}

void FreeFormulaListItem(FormulaList *formula_list_item)
{
	FreeFormula(formula_list_item->formula);
	free(formula_list_item);
}

void FreeFormulaList(FormulaList *formula_list)
{
	FormulaList *tmp;
	while(formula_list)
	{
		tmp = formula_list;
		formula_list = formula_list->next;
		FreeFormulaListItem(tmp);
	}
}

FormulaList *MakeFormulaListItem(Formula *formula, int ID)
{
	FormulaList *new_list_item = malloc(sizeof(FormulaList));
	if(new_list_item == NULL)
		return NULL;
	new_list_item->formula = formula;
	new_list_item->ID = ID;
	new_list_item->next = NULL;
	return new_list_item;
}

/********************************* Parsing **************************/
static FormulaTokenType GetNextToken(const char *formula,
									 int len, int *token_len)
{
	static char *string_array[] = {"MAX(", "MIN(", "WHERE(",
		"ALL=", "ALL==", "ALL!=" , "ALL<", "ALL<=", "ALL>", "ALL>=",
		"ANY=", "ANY==", "ANY!=" , "ANY<", "ANY<=", "ANY>", "ANY>=",
		NULL};
	static FormulaTokenType token_array[] = {MAXToken, MINToken, WHEREToken,
		All_EqualToken, All_EqualToken,	All_NotEqToken,
		All_LessToken, All_LessEqToken, All_MoreToken, All_MoreEqToken,
		Any_EqualToken, Any_EqualToken,	Any_NotEqToken,
		Any_LessToken, Any_LessEqToken, Any_MoreToken, Any_MoreEqToken,
		EmptyToken};

	int i = 0, j;
	
	while((formula[i] == ' ') && (i < len))
		i++;
	if(i == len)
	{
		*token_len = i;
		return EmptyToken;
	}
	
	for(j = 0; string_array[j]; j++)
		if(strncmp(&formula[i], string_array[j], strlen(string_array[j])) == 0)
		{
			*token_len = i + strlen(string_array[j]);
			return token_array[j];
		}
	switch(formula[i])
	{
	case '+':
		*token_len = i + 1;
		return PlusToken;
	case '-':
		*token_len = i + 1;
		return MinusToken;
	case '*':
		*token_len = i + 1;
		return MultToken;
	case '/':
		*token_len = i + 1;
		return DivToken;
	case '=':
		if((i + 1 < len) && (formula[i + 1] == '='))
		{
			*token_len = i + 2;
			return EqualToken;
		}
		*token_len = i + 1;
		return AssignToken;
	case '>':
		if((i + 1 < len) && (formula[i + 1] == '='))
		{
			*token_len = i + 2;
			return MoreEqToken;
		}
		*token_len = i + 1;
		return MoreToken;
	case '<':
		if((i + 1 < len) && (formula[i + 1] == '='))
		{
			*token_len = i + 2;
			return LessEqToken;
		}
		*token_len = i + 1;
		return LessToken;
	case '!':
		if((i + 1 < len) && (formula[i + 1] == '='))
		{
			*token_len = i + 2;
			return NotEqToken;
		}
		*token_len = i + 1;
		return NotToken;
	case '&':
		if((i + 1 < len) && (formula[i + 1] == '&'))
		{
			*token_len = i + 2;
			return AndToken;
		}
		*token_len = i;
		return ErrorToken;
	case '|':
		if((i + 1 < len) && (formula[i + 1] == '|'))
		{
			*token_len = i + 2;
			return OrToken;
		}
		*token_len = i;
		return ErrorToken;
	case '.':
		if((i + 1 < len) && (formula[i + 1] == '.'))
		{
			*token_len = i + 2;
			return RangeToken;
		}
		*token_len = i;
		return ErrorToken;
	case '[':
		*token_len = i + 1;
		return IndexStartToken;
	case ']':
		*token_len = i + 1;
		return IndexEndToken;
	case '(':
		*token_len = i + 1;
		return ParenthesisStartToken;
	case ')':
		*token_len = i + 1;
		return ParenthesisEndToken;
	case ',':
		*token_len = i + 1;
		return CommaToken;
	case ';':
		*token_len = i + 1;
		return SemicolonToken;
	}
	if(isdigit(formula[i]))
	{
		FormulaTokenType type = IntegerToken;
		i++;
		while(i < len)
		{
			switch(formula[i])
			{
			case '.':
				if((i + 1 < len) && (formula[i + 1] == '.')) /* it is range token */
				{
					*token_len = i;
					return type;
				}
				if(type == IntegerToken)
				{
					type = RealToken;
					break;
				}
				*token_len = i;
				return ErrorToken;
			case 'D':
				*token_len = i + 1;
				return DoubleToken;
			case 'R':
				*token_len = i + 1;
				return RealToken;
			default:
				if(!isdigit(formula[i]))
				{
					*token_len = i;
					return type;
				}
			}
			i++;
		}
		*token_len = i;
		return type;
	}
	if(isalnum(formula[i]))
	{
		i++;
		while((i < len) &&
			(isalnum(formula[i]) || isdigit(formula[i]) || 
			(formula[i] == '_')))
			i++;
		*token_len = i;
		return VariableToken;
	}
	*token_len = i;
	return ErrorToken;
}

static int GetCorrespondedParenthesisEnd(const char *formula, int len)
{
	int i = 0, counter = 1;

	while(i < len)
	{
		if(formula[i] == '(')
			counter++;
		else if(formula[i] == ')')
		{
			if(--counter == 0)
				return i;
		}
		i++;
	}
	return 0;
}

/********************************* Array Functions ******************/
static void ArrayInit(FormulaArray *ar)
{
	ar->amount = 0;
}
static void ArrayFree(FormulaArray *ar)
{
	int i;
	for(i = 0; i < ar->amount; i++)
		FreeFormula(ar->formula[i]);
	ar->amount = 0;
}
static Formula* ArrayCreateItem(FormulaArray *ar)
{
	Formula *new_formula;
	if(ar->amount == MAX_FORMULA_ITEMS)
	{
		assert(FALSE);
		return NULL;
	}
	if((new_formula = malloc(sizeof(Formula))) == NULL)
		return NULL;
	new_formula->type = NoItem;
	ar->formula[ar->amount++] = new_formula;
	return new_formula;
}

static void ArrayExclude(FormulaArray *ar, int number)
{
	int i;

	assert(number < ar->amount);
	assert(number >= 0);

	for(i = number; i < ar->amount - 1; i++)
		ar->formula[i] = ar->formula[i + 1];
	ar->amount--;
}
static void ArrayReplaceLast(FormulaArray *ar, Formula *f)
{
	if(ar->amount == 0)
	{
		assert(FALSE);
	}
	else
		FreeFormula(ar->formula[--(ar->amount)]);
	ar->formula[ar->amount++] = f;
}

static BOOL ArrayCheckNotOperator(FormulaArray *ar)
{
	if(ar->amount < 2)
		return(FALSE);
	if((ar->formula[ar->amount - 2]->type == OperatorItem) &&
		(ar->formula[ar->amount - 2]->item.noperator->left == NULL))
		return(FALSE);
	return(TRUE);
}

static BOOL ArrayCheckFirstOrOperator(FormulaArray *ar)
{
	if(ar->amount < 2)
		return(TRUE);
	if((ar->formula[ar->amount - 2]->type == OperatorItem) &&
		(ar->formula[ar->amount - 2]->item.noperator->left == NULL))
		return(TRUE);
	return(FALSE);
}

/*********************************  ******************/

static OperatorType Token2Operator(FormulaTokenType token)
{
	switch(token)
	{
	case PlusToken:
		return PLUS;
	case MinusToken:
		return MINUS;
	case DivToken:
		return DIV;
	case MultToken:
		return MULT;
	case EqualToken:
	case AssignToken:
		return EQUAL;
	case NotEqToken:
		return NEQUAL;
	case MoreToken:
		return MORE;
	case MoreEqToken:
		return MOREEQ;
	case LessToken:
		return LESS;
	case LessEqToken:
		return LESSEQ;
	case All_EqualToken:
		return ALL_EQUAL;
	case All_NotEqToken:
		return ALL_NEQUAL;
	case All_MoreToken:
		return ALL_MORE;
	case All_MoreEqToken:
		return ALL_MOREEQ;
	case All_LessToken:
		return ALL_LESS;
	case All_LessEqToken:
		return ALL_LESSEQ;
	case Any_EqualToken:
		return ANY_EQUAL;
	case Any_NotEqToken:
		return ANY_NEQUAL;
	case Any_MoreToken:
		return ANY_MORE;
	case Any_MoreEqToken:
		return ANY_MOREEQ;
	case Any_LessToken:
		return ANY_LESS;
	case Any_LessEqToken:
		return ANY_LESSEQ;
	case NotToken:
		return NOT;
	case AndToken:
		return AND;
	case OrToken:
		return OR;
	}
	assert(FALSE);
	return OR;
}

static const char *Operator2Char(OperatorType type)
{
	static char *operators_char[] = {"+", "-", "/", "*", "==", "!=",
		">", ">=", "<", "<=",
		"ALL==", "ALL!=", "ALL>", "ALL>=", "ALL<", "ALL<=",
		"ANY==", "ANY!=", "ANY>", "ANY>=", "ANY<", "ANY<=",
		"!", "&&", "||"};

	return operators_char[type];
}

static BOOL IsOperatorAllowArray(OperatorType type)
{
	return ((type >= ALL_EQUAL) && (type <= ANY_LESSEQ));
}

static OperatorType ReverseOperatorType(OperatorType type)
{
	switch(type)
	{
	case ALL_MORE:
		return ALL_LESS;
	case ALL_MOREEQ:
		return ALL_LESSEQ;
	case ALL_LESS:
		return ALL_MORE;
	case ALL_LESSEQ:
		return ALL_MOREEQ;
	case ANY_MORE:
		return ANY_LESS;
	case ANY_MOREEQ:
		return ANY_LESSEQ;
	case ANY_LESS:
		return ANY_MORE;
	case ANY_LESSEQ:
		return ANY_MOREEQ;
	}
	return type;
}

static BOOL IsCurrentOperatorLevel(OperatorType type, int level)
{
	switch(level)
	{
	case 0:
		if(type == NOT)
			return TRUE;
		break;
	case 1:
		switch(type)
		{
		case DIV:
		case MULT:
		case AND:
			return TRUE;
		}
		break;
	case 2:
		switch(type)
		{
		case PLUS:
		case MINUS:
		case OR:
			return TRUE;
		}
		break;
	case 3:
		switch(type)
		{
		case EQUAL:
		case NEQUAL:
		case MORE:
		case MOREEQ:
		case LESS:
		case LESSEQ:
		case ALL_EQUAL:
		case ALL_NEQUAL:
		case ALL_MORE:
		case ALL_MOREEQ:
		case ALL_LESS:
		case ALL_LESSEQ:
		case ANY_EQUAL:
		case ANY_NEQUAL:
		case ANY_MORE:
		case ANY_MOREEQ:
		case ANY_LESS:
		case ANY_LESSEQ:
			return TRUE;
		}
		break;
	}
	return FALSE;
}

/************** Build Formule from the text notation **************/
Formula* BuildFormula(const char *formula, int len, int *error_pos, char **error_message)
{
	int i = 0, j, token_len, item_len, val, norma_index;
	FormulaTokenType token;
	Formula *new_formula;
	/* we parse char record onto several formulas in this array
		and then will build one formula looking on priorities of operations */
	FormulaArray ar;
	Index *new_index, *last_index;

	*error_message = NULL;
	ArrayInit(&ar);

	while(i < len)
	{
		if(formula[i] == ' ')
		{
			i++;
			continue;
		}
		if((new_formula = ArrayCreateItem(&ar)) == NULL)
		{
			*error_pos = 0;
			*error_message = OUT_MEMORY;
			ArrayFree(&ar);
			return NULL;
		}

		token = GetNextToken(&formula[i], len - i, &token_len);
		switch(token)
		{
		case EmptyToken:
			break;
		case ErrorToken:
		case RangeToken:
		case CommaToken:
		case SemicolonToken:
		case ParenthesisEndToken:
		case IndexStartToken:
		case IndexEndToken:
		default:
			*error_pos = i;
			*error_message = SYNTAX_ERROR;
			ArrayFree(&ar);
			return NULL;
		case MAXToken:
		case MINToken:
		case WHEREToken:
		case ParenthesisStartToken:
		{
			Formula *subformula;

			if(!ArrayCheckFirstOrOperator(&ar))
			{
				*error_pos = i;
				*error_message = SYNTAX_ERROR;
				ArrayFree(&ar);
				return NULL;
			}
			i += token_len;
			item_len = GetCorrespondedParenthesisEnd(
				&formula[i], len - i);
			if(item_len == 0)
			{
				*error_pos = i;
				*error_message = NO_CORRESPONDED_PARENTHESIS_END;
				ArrayFree(&ar);
				return NULL;
			}
			if((subformula = BuildFormula(&formula[i], 
				item_len, error_pos, error_message)) == NULL)
			{
				*error_pos += i;
				ArrayFree(&ar);
				return NULL;
			}

			if(token == ParenthesisStartToken)
			{
				ArrayReplaceLast(&ar, subformula);
				i += item_len + 1; /* +1 because ')' */
				break;
			}
			
			if(((subformula->type != VariableItem) &&
				(subformula->type != FunctionItem)) ||
				((subformula->type == FunctionItem) &&
				((subformula->item.function->type != WHEREFunction) ||
				(token == WHEREToken))))
			{
				*error_pos = i;
				*error_message = SYNTAX_ERROR;
				ArrayFree(&ar);
				FreeFormula(subformula);
				return NULL;
			}

			if((new_formula->item.function = malloc(sizeof(Function))) == NULL)
			{
				*error_pos = 0;
				*error_message = OUT_MEMORY;
				ArrayFree(&ar);
				FreeFormula(subformula);
				return NULL;
			}
			new_formula->type = FunctionItem;
			new_formula->item.function->operand = subformula;
			if(token == MAXToken)
				new_formula->item.function->type = MAXFunction;
			else if(token == MINToken)
				new_formula->item.function->type = MINFunction;
			else
				new_formula->item.function->type = WHEREFunction;

			i += item_len + 1; /* +1 because ')' */
			break;
		}

		case PlusToken:
		case MultToken:
		case DivToken:
		case EqualToken:
		case AssignToken:
		case NotEqToken:
		case MoreEqToken:
		case MoreToken:
		case LessEqToken:
		case LessToken:
		case All_EqualToken:
		case All_NotEqToken:
		case All_MoreEqToken:
		case All_MoreToken:
		case All_LessEqToken:
		case All_LessToken:
		case Any_EqualToken:
		case Any_NotEqToken:
		case Any_MoreEqToken:
		case Any_MoreToken:
		case Any_LessEqToken:
		case Any_LessToken:
		case AndToken:
		case OrToken:
			if(!ArrayCheckNotOperator(&ar))
			{
				*error_pos = i;
				*error_message = SYNTAX_ERROR;
				ArrayFree(&ar);
				return NULL;
			}
		case NotToken:
		case MinusToken:
			if((new_formula->item.noperator = malloc(sizeof(Operator))) == NULL)
			{
				*error_pos = 0;
				*error_message = OUT_MEMORY;
				ArrayFree(&ar);
				return NULL;
			}
			new_formula->item.noperator->left = NULL;
			new_formula->item.noperator->right = NULL;
			new_formula->item.noperator->type = Token2Operator(token);
			new_formula->type = OperatorItem;
			i += token_len;
			break;

		case IntegerToken:
		case RealToken:
		case DoubleToken:
			if(!ArrayCheckFirstOrOperator(&ar))
			{
				*error_pos = i;
				*error_message = SYNTAX_ERROR;
				ArrayFree(&ar);
				return NULL;
			}
			if((new_formula->item.value = MakeValue()) == NULL)
			{
				*error_pos = 0;
				*error_message = OUT_MEMORY;
				ArrayFree(&ar);
				return NULL;
			}

			switch(token)
			{
			case IntegerToken:
				new_formula->item.value->type = INTEGER_T;
				new_formula->item.value->value.int_value = atoi(&formula[i]);
				break;
			case RealToken:
				new_formula->item.value->type = REAL_T;
				new_formula->item.value->value.real_value = (float)atof(&formula[i]);
				break;
			case DoubleToken:
				new_formula->item.value->type = DOUBLE_T;
				new_formula->item.value->value.double_value = atof(&formula[i]);
				break;
			}
			new_formula->type = ValueItem;
			i += token_len;
			break;
		case VariableToken:
			if(!ArrayCheckFirstOrOperator(&ar))
			{
				*error_pos = i;
				*error_message = SYNTAX_ERROR;
				ArrayFree(&ar);
				return NULL;
			}
			if(token_len >= MAX_VARIABLE_NAME)
			{
				*error_pos = i + MAX_VARIABLE_NAME;
				*error_message = SYNTAX_ERROR;
				ArrayFree(&ar);
				return NULL;
			}
			if((new_formula->item.variable = malloc(sizeof(Variable))) == NULL)
			{
				*error_pos = 0;
				*error_message = OUT_MEMORY;
				ArrayFree(&ar);
				return NULL;
			}
			memcpy(new_formula->item.variable->name, &formula[i], token_len);
			new_formula->item.variable->name[token_len] = 0;
			new_formula->item.variable->index = NULL;
			new_formula->type = VariableItem;
			i += token_len;
			if(i >=len )
				break;
			switch(GetNextToken(&formula[i], len - i, &token_len))
			{
			case IndexStartToken: /* index start from '[' - NORMA style */
				norma_index = 1;
				break;
			case ParenthesisStartToken: /* index start from '(' - Fortran style */
				norma_index = 0;
				break;
			default:
				norma_index = -1;
				break;
			}
			if(norma_index == -1)
				break;

			i += token_len;
			/* Index parsing */
			new_index = last_index = NULL;
			while((i < len) && (token != IndexEndToken) && (token != ParenthesisEndToken))
			{
				if(formula[i] == ' ')
				{
					i++;
					continue;
				}
				token = GetNextToken(&formula[i], len - i, &token_len);
				switch(token)
				{
				case IntegerToken:
					val = atoi(&formula[i]);
					if(new_index == NULL)
					{
						if(norma_index)
						{ /* error: Norma index must has an index name */ 
							*error_pos = i;
							*error_message = SYNTAX_ERROR;
							ArrayFree(&ar);
							return NULL;
						}
						/* new Fortran index */
						if((new_index = malloc(sizeof(Index))) == NULL)
						{
							*error_pos = 0;
							*error_message = OUT_MEMORY;
							ArrayFree(&ar);
							return NULL;
						}
						new_index->name[0] = 0;
						new_index->next = NULL;
						new_index->found = FALSE;
						new_index->from = new_index->to = val;
						if(last_index)
							last_index->next = new_index;
						else
							new_formula->item.variable->index = new_index;
						last_index = new_index;
					}
					else
					{
						if(new_index->from == -1) /* it was the index name (variable) before */
							new_index->from = new_index->to = val;
						else if(new_index->to == -1) /* it was the range */
								new_index->to = val;
						else
						{
							*error_pos = i;
							*error_message = SYNTAX_ERROR;
							ArrayFree(&ar);
							return NULL;
						}
					}
					i += token_len;
					break;
				case VariableToken:
					if((token_len >= MAX_VARIABLE_NAME) ||
						(new_index != NULL))
					{
						*error_pos = i + 
							((token_len >= MAX_VARIABLE_NAME) ? MAX_VARIABLE_NAME : 0);
						*error_message = SYNTAX_ERROR;
						ArrayFree(&ar);
						return NULL;
					}
					if((new_index = malloc(sizeof(Index))) == NULL)
					{
						*error_pos = 0;
						*error_message = OUT_MEMORY;
						ArrayFree(&ar);
						return NULL;
					}
					memcpy(new_index->name, &formula[i], token_len);
					new_index->name[token_len] = 0;
					new_index->next = NULL;
					new_index->found = FALSE;
					new_index->from = new_index->to = -1; /* set that it was the index name (variable) */
					if(last_index)
						last_index->next = new_index;
					else
						new_formula->item.variable->index = new_index;
					last_index = new_index;
					i += token_len;
					break;
				case AssignToken:
					if((new_index == NULL) || (new_index->from != -1))
					{
						*error_pos = i;
						*error_message = SYNTAX_ERROR;
						ArrayFree(&ar);
						return NULL;
					}
					i += token_len;
					break;
				case RangeToken:
					if(new_index == NULL)
					{
						*error_pos = i;
						*error_message = SYNTAX_ERROR;
						ArrayFree(&ar);
						return NULL;
					}
					new_index->to = -1;
					i += token_len;
					break;
				case CommaToken:
				case SemicolonToken:
					if((new_index == NULL) ||
						(new_index->from == -1) ||
						(new_index->to == -1))
					{
						*error_pos = i;
						*error_message = SYNTAX_ERROR;
						ArrayFree(&ar);
						return NULL;
					}
					i += token_len;
					new_index = NULL;
					break;
				case IndexEndToken:
				case ParenthesisEndToken:
					if((new_index == NULL) ||
						(new_index->from == -1) ||
						(new_index->to == -1))
					{
						*error_pos = i;
						*error_message = SYNTAX_ERROR;
						ArrayFree(&ar);
						return NULL;
					}
					i += token_len;
					break;
				default:
					*error_pos = i;
					*error_message = SYNTAX_ERROR;
					ArrayFree(&ar);
					return NULL;
				}
			}
			break; /* from case VariableToken: */
		} /* from main switch */
	} /* from while(i < len) */

	if(ar.amount == 0)
		return NULL;
	if(ar.formula[ar.amount - 1]->type == NoItem)
	{
		FreeFormula(ar.formula[ar.amount - 1]);
		ar.amount--;
	}
	if(ar.amount == 0)
		return NULL;
	if((ar.formula[ar.amount - 1]->type == OperatorItem) &&
				(ar.formula[ar.amount - 1]->item.noperator->left == NULL))
	{ /* we can't have operator at the end */
		*error_pos = len;
		*error_message = SYNTAX_ERROR;
		ArrayFree(&ar);
		return NULL;
	}

	for(i = 0; i < 4; i++)
	{
		for(j = 0; j < ar.amount; j++)
		{
			if((ar.formula[j]->type == OperatorItem) &&
				(ar.formula[j]->item.noperator->left == NULL) &&
				IsCurrentOperatorLevel(ar.formula[j]->item.noperator->type, i))
			{
				assert((j + 1 < ar.amount) && ((ar.formula[j + 1]->type != OperatorItem) ||
					(ar.formula[j + 1]->item.noperator->left != NULL)));

				if((ar.formula[j]->item.noperator->type == NOT) ||
					((ar.formula[j]->item.noperator->type == MINUS) && (j == 0)))
				{ /* unary operator */
					ar.formula[j]->item.noperator->left = ar.formula[j + 1];
					ArrayExclude(&ar, j + 1);
				}
				else
				{ /* binary operator */
					assert((j > 0) && ((ar.formula[j - 1]->type != OperatorItem) ||
						(ar.formula[j - 1]->item.noperator->left != NULL)));

					ar.formula[j]->item.noperator->left = ar.formula[j - 1];
					ArrayExclude(&ar, j - 1);
					j--;
					ar.formula[j]->item.noperator->right = ar.formula[j + 1];
					ArrayExclude(&ar, j + 1);
				}
			}
		}
	}

	assert(ar.amount == 1);
	return ar.formula[0];
}

/************** Calculate Formula **************/

static void TypeCast2Double(Value *value)
{
	if(value->type != DOUBLE_T)
	{
		value->value.double_value = (double) (value->type == INTEGER_T ? value->value.int_value : value->value.real_value);
		value->type = DOUBLE_T;
	}
}

static void TypeCast2Real(Value *value)
{
	if(value->type != REAL_T)
	{
		value->value.real_value = (float) (value->type == INTEGER_T ? value->value.int_value : value->value.double_value);
		value->type = REAL_T;
	}
}

static void TypeCast(Value *left, int amount_l, Value *right)
{
	int i;

	if(right == NULL)
		return;
	if(left->type == DOUBLE_T)
	{
		TypeCast2Double(right);
		return;
	}
	if(right->type == DOUBLE_T)
	{
		for(i = 0; i < amount_l; i++) 
			TypeCast2Double(&left[i]);
		return;
	}
	if(left->type == REAL_T)
	{
		TypeCast2Real(right);
		return;
	}
	if(right->type == REAL_T)
	{
		for(i = 0; i < amount_l; i++) 
			TypeCast2Real(&left[i]);
		return;
	}
}

/*-------------------------------------------------------*/
BOOL IsValueTrue(const Value *value)
{
	switch(value->type)
	{
	case INTEGER_T:
		return (value->value.int_value != 0);
	case REAL_T:
		return (value->value.real_value != 0);
	case DOUBLE_T:
		return (value->value.double_value != 0);
	}
	return FALSE;
}
/*-------------------------------------------------------*/
BOOL IsAnyValueTrue(const ValuesProcArray *value)
{
	int i, j;
	for(i = 0; i < value->proc_amount; i++)
		for(j = 0; j < value->array[i].amount; j++)
			if(IsValueTrue(&(value->array[i].values[j])))
				return TRUE;
	return FALSE;
}

/*-------------------------------------------------------*/
ValuesProcArray *CalculateFormula(const Formula *formula, char **error_message)
{
	ValuesProcArray *left_array, *right_array;
	int proc_num;
	BOOL change, copy;
	static char msg[200];
	OperatorType operator_type;

	switch(formula->type)
	{
	case VariableItem:
#ifdef DEBUG
		fprintf(stderr, "CalculateFormula: VariableItem before\n");
#endif
		if((left_array = CalculateVariable(formula->item.variable, error_message, FALSE)) == NULL)
			return NULL;
#ifdef DEBUG
		fprintf(stderr, "CalculateFormula: VariableItem, left_array == %p\n", left_array);
#endif
		return left_array;
	case OperatorItem:
		if((left_array = CalculateFormula(formula->item.noperator->left, error_message)) == NULL)
		{
#ifdef DEBUG
			fprintf(stderr, "CalculateFormula: OperatorItem, left_array == NULL\n");
#endif
			return NULL;
		}
		if(formula->item.noperator->right) /* binary operator */
		{
			if((right_array = CalculateFormula(formula->item.noperator->right, error_message)) == NULL)
			{
#ifdef DEBUG
				fprintf(stderr, "CalculateFormula: OperatorItem, right_array == NULL\n");
#endif
				FreeValuesProcArray(left_array);
				return NULL;
			}
		}
		else
			right_array = NULL;

		operator_type = formula->item.noperator->type;

		if(right_array && (left_array->proc_amount != right_array->proc_amount) &&
			(left_array->proc_amount > 1) && (right_array->proc_amount > 1))
		{
			sprintf(msg, "Different processor amount in operator %s",
				Operator2Char(operator_type));
			*error_message = msg;
			FreeValuesProcArray(left_array);
			FreeValuesProcArray(right_array);
			return NULL;
		}

		if(right_array && (right_array->proc_amount > left_array->proc_amount))
		{
			ValuesProcArray *tmp = left_array;
			left_array = right_array;
			right_array = tmp;
		}

		for(proc_num = 0; proc_num < left_array->proc_amount; proc_num++)
		{
			Value *left, *right, *rez_v;
			int i, rez, amount_l, amount_r;

			if((amount_l = left_array->array[proc_num].amount) == 0)
				continue; // No values => nothing to do
			left = left_array->array[proc_num].values;
			if(right_array)
			{
				int r_proc_num = proc_num < right_array->proc_amount ? proc_num : right_array->proc_amount - 1;
				right = right_array->array[r_proc_num].values;
				amount_r = right_array->array[r_proc_num].amount;
			}
			else
			{
				right = NULL;
				amount_r = 0;
			}

			if(((amount_l > 1) && (amount_r > 1))||/* No one operators allow this */
				(((amount_l > 1) || (amount_r > 1)) &&
				!IsOperatorAllowArray(operator_type)))
			{
				if(IsOperatorAllowArray(operator_type))
					sprintf(msg, "Operator %s can't have the arrays on the both left and right sides",
						Operator2Char(operator_type));
				else
					sprintf(msg, "Operator %s isn't applicable  for arrays",
						Operator2Char(operator_type));
				*error_message = msg;
				FreeValuesProcArray(left_array);
				FreeValuesProcArray(right_array);
				return NULL;
			}
				
			if(amount_r > 1)/* we have the array on the right side */
			{
				Value *tmp_v;
				int tmp_i;

				assert(proc_num < right_array->proc_amount);

				tmp_v = left;
				tmp_i = amount_l;
				left = right;
				amount_l = amount_r;
				right = tmp_v;
				amount_r = tmp_i;
				operator_type = ReverseOperatorType(operator_type);
			}
			assert(amount_r <= 1);

			TypeCast(left, amount_l, right);

			switch(operator_type)
			{
			case PLUS:
				switch(left->type)
				{
				case INTEGER_T:
					left->value.int_value += right->value.int_value;
					break;
				case REAL_T:
					left->value.real_value += right->value.real_value;
					break;
				case DOUBLE_T:
					left->value.double_value += right->value.double_value;
					break;
				default:
					assert(FALSE);
				}
				break;

			case MINUS:
				if(right == NULL) /* unary - */
					switch(left->type)
					{
					case INTEGER_T:
						left->value.int_value = -left->value.int_value;
						break;
					case REAL_T:
						left->value.real_value = -left->value.real_value;
						break;
					case DOUBLE_T:
						left->value.double_value = -left->value.double_value;
						break;
					default:
						assert(FALSE);
					}
				else
					switch(left->type)
					{
					case INTEGER_T:
						left->value.int_value -= right->value.int_value;
						break;
					case REAL_T:
						left->value.real_value -= right->value.real_value;
						break;
					case DOUBLE_T:
						left->value.double_value -= right->value.double_value;
						break;
					default:
						assert(FALSE);
					}
				break;

			case DIV:
				switch(left->type)
				{
				case INTEGER_T:
					if(right->value.int_value == 0)
					{
						*error_message = DIVIDE_BY_ZERO;
						return NULL;
					}
					left->value.int_value /= right->value.int_value;
					break;
				case REAL_T:
					if(right->value.real_value == 0)
					{
						*error_message = DIVIDE_BY_ZERO;
						return NULL;
					}
					left->value.real_value /= right->value.real_value;
					break;
				case DOUBLE_T:
					if(right->value.double_value == 0)
					{
						*error_message = DIVIDE_BY_ZERO;
						return NULL;
					}
					left->value.double_value /= right->value.double_value;
					break;
				default:
					assert(FALSE);
				}
				break;

			case MULT:
				switch(left->type)
				{
				case INTEGER_T:
					left->value.int_value *= right->value.int_value;
					break;
				case REAL_T:
					left->value.real_value *= right->value.real_value;
					break;
				case DOUBLE_T:
					left->value.double_value *= right->value.double_value;
					break;
				default:
					assert(FALSE);
				}
				break;

			case NEQUAL:
			case EQUAL:
				switch(left->type)
				{
				case INTEGER_T:
					if(left->value.int_value == right->value.int_value)
						left->value.int_value = 1;
					else
						left->value.int_value = 0;
					break;
				case REAL_T:
					if(left->value.real_value == right->value.real_value)
						left->value.int_value = 1;
					else
						left->value.int_value = 0;
					break;
				case DOUBLE_T:
					if(left->value.double_value == right->value.double_value)
						left->value.int_value = 1;
					else
						left->value.int_value = 0;
					break;
				default:
					assert(FALSE);
				}
				left->type = INTEGER_T;
				if(operator_type == NEQUAL)
					left->value.int_value = !left->value.int_value;
				break;

			case LESSEQ:
			case MORE:
				switch(left->type)
				{
				case INTEGER_T:
					if(left->value.int_value > right->value.int_value)
						left->value.int_value = 1;
					else
						left->value.int_value = 0;
					break;
				case REAL_T:
					if(left->value.real_value > right->value.real_value)
						left->value.int_value = 1;
					else
						left->value.int_value = 0;
					break;
				case DOUBLE_T:
					if(left->value.double_value > right->value.double_value)
						left->value.int_value = 1;
					else
						left->value.int_value = 0;
					break;
				default:
					assert(FALSE);
				}
				left->type = INTEGER_T;
				if(operator_type == LESSEQ)
					left->value.int_value = !left->value.int_value;
				break;

			case LESS:
			case MOREEQ:
				switch(left->type)
				{
				case INTEGER_T:
					if(left->value.int_value >= right->value.int_value)
						left->value.int_value = 1;
					else
						left->value.int_value = 0;
					break;
				case REAL_T:
					if(left->value.real_value >= right->value.real_value)
						left->value.int_value = 1;
					else
						left->value.int_value = 0;
					break;
				case DOUBLE_T:
					if(left->value.double_value >= right->value.double_value)
						left->value.int_value = 1;
					else
						left->value.int_value = 0;
					break;
				default:
					assert(FALSE);
				}
				left->type = INTEGER_T;
				if(operator_type == LESS)
					left->value.int_value = !left->value.int_value;
				break;

			case ALL_EQUAL:
			case ALL_NEQUAL:
			case ALL_MORE:
			case ALL_LESSEQ:
			case ALL_MOREEQ:
			case ALL_LESS:
				rez = 1;
				for(i = 0; ((i < amount_l) && rez); i++)
					switch(operator_type)
					{
					case ALL_EQUAL:
						switch(left->type)
						{
						case INTEGER_T:
							if(left[i].value.int_value != right->value.int_value)
								rez = 0;
							break;
						case REAL_T:
							if(left[i].value.real_value != right->value.real_value)
								rez = 0;
							break;
						case DOUBLE_T:
							if(left[i].value.double_value != right->value.double_value)
								rez = 0;
							break;
						default:
							assert(FALSE);
						}
						break;

					case ALL_NEQUAL:
						switch(left->type)
						{
						case INTEGER_T:
							if(left[i].value.int_value == right->value.int_value)
								rez = 0;
							break;
						case REAL_T:
							if(left[i].value.real_value == right->value.real_value)
								rez = 0;
							break;
						case DOUBLE_T:
							if(left[i].value.double_value == right->value.double_value)
								rez = 0;
							break;
						default:
							assert(FALSE);
						}
						break;

					case ALL_MORE:
						switch(left->type)
						{
						case INTEGER_T:
							if(left[i].value.int_value <= right->value.int_value)
								rez = 0;
							break;
						case REAL_T:
							if(left[i].value.real_value <= right->value.real_value)
								rez = 0;
							break;
						case DOUBLE_T:
							if(left[i].value.double_value <= right->value.double_value)
								rez = 0;
							break;
						default:
							assert(FALSE);
						}
						break;

					case ALL_LESSEQ:
						switch(left->type)
						{
						case INTEGER_T:
							if(left[i].value.int_value > right->value.int_value)
								rez = 0;
							break;
						case REAL_T:
							if(left[i].value.real_value > right->value.real_value)
								rez = 0;
							break;
						case DOUBLE_T:
							if(left[i].value.double_value > right->value.double_value)
								rez = 0;
							break;
						default:
							assert(FALSE);
						}
						break;

					case ALL_MOREEQ:
						switch(left->type)
						{
						case INTEGER_T:
							if(left[i].value.int_value < right->value.int_value)
								rez = 0;
							break;
						case REAL_T:
							if(left[i].value.real_value < right->value.real_value)
								rez = 0;
							break;
						case DOUBLE_T:
							if(left[i].value.double_value < right->value.double_value)
								rez = 0;
							break;
						default:
							assert(FALSE);
						}
						break;

					case ALL_LESS:
						switch(left->type)
						{
						case INTEGER_T:
							if(left[i].value.int_value >= right->value.int_value)
								rez = 0;
							break;
						case REAL_T:
							if(left[i].value.real_value >= right->value.real_value)
								rez = 0;
							break;
						case DOUBLE_T:
							if(left[i].value.double_value >= right->value.double_value)
								rez = 0;
							break;
						default:
							assert(FALSE);
						}
						break;

					default:
						assert(FALSE);
					}
				FreeValues(left, amount_l);
				if((left = MakeValue()) == NULL)
				{
					*error_message = OUT_MEMORY;
					return NULL;
				}
				left->type = INTEGER_T;
				left->value.int_value = rez;
				left_array->array[proc_num].values = left;
				left_array->array[proc_num].amount = 1;
				break;

			case ANY_EQUAL:
			case ANY_NEQUAL:
			case ANY_MORE:
			case ANY_LESSEQ:
			case ANY_MOREEQ:
			case ANY_LESS:
				rez = 0;
				if((rez_v = MakeValues((amount_l > MAX_DISPLAYED_VALUES ? MAX_DISPLAYED_VALUES : amount_l))) == NULL)
				{
					*error_message = OUT_MEMORY;
					return NULL;
				}
				copy = FALSE;
				for(i = 0; (i < amount_l) && (rez < MAX_DISPLAYED_VALUES); i++)
				{
					switch(operator_type)
					{
					case ANY_EQUAL:
						switch(left->type)
						{
						case INTEGER_T:
							if(left[i].value.int_value == right->value.int_value)
								copy = TRUE;;
							break;
						case REAL_T:
							if(left[i].value.real_value == right->value.real_value)
								copy = TRUE;;
							break;
						case DOUBLE_T:
							if(left[i].value.double_value == right->value.double_value)
								copy = TRUE;;
							break;
						default:
							assert(FALSE);
						}
						break;
					
					case ANY_NEQUAL:
						switch(left->type)
						{
						case INTEGER_T:
							if(left[i].value.int_value != right->value.int_value)
								copy = TRUE;;
							break;
						case REAL_T:
							if(left[i].value.real_value != right->value.real_value)
								copy = TRUE;;
							break;
						case DOUBLE_T:
							if(left[i].value.double_value != right->value.double_value)
								copy = TRUE;;
							break;
						default:
							assert(FALSE);
						}
						break;
					
					case ANY_MORE:
						switch(left->type)
						{
						case INTEGER_T:
							if(left[i].value.int_value > right->value.int_value)
								copy = TRUE;;
							break;
						case REAL_T:
							if(left[i].value.real_value > right->value.real_value)
								copy = TRUE;;
							break;
						case DOUBLE_T:
							if(left[i].value.double_value > right->value.double_value)
								copy = TRUE;;
							break;
						default:
							assert(FALSE);
						}
						break;
					
					case ANY_MOREEQ:
						switch(left->type)
						{
						case INTEGER_T:
							if(left[i].value.int_value >= right->value.int_value)
								copy = TRUE;;
							break;
						case REAL_T:
							if(left[i].value.real_value >= right->value.real_value)
								copy = TRUE;;
							break;
						case DOUBLE_T:
							if(left[i].value.double_value >= right->value.double_value)
								copy = TRUE;;
							break;
						default:
							assert(FALSE);
						}
						break;
					
					case ANY_LESS:
						switch(left->type)
						{
						case INTEGER_T:
							if(left[i].value.int_value < right->value.int_value)
								copy = TRUE;;
							break;
						case REAL_T:
							if(left[i].value.real_value < right->value.real_value)
								copy = TRUE;;
							break;
						case DOUBLE_T:
							if(left[i].value.double_value < right->value.double_value)
								copy = TRUE;;
							break;
						default:
							assert(FALSE);
						}
						break;
					
					case ANY_LESSEQ:
						switch(left->type)
						{
						case INTEGER_T:
							if(left[i].value.int_value <= right->value.int_value)
								copy = TRUE;;
							break;
						case REAL_T:
							if(left[i].value.real_value <= right->value.real_value)
								copy = TRUE;;
							break;
						case DOUBLE_T:
							if(left[i].value.double_value <= right->value.double_value)
								copy = TRUE;;
							break;
						default:
							assert(FALSE);
						}
						break;
					
					default:
						assert(FALSE);
					}

					if(copy)
					{
						CopyRIValue(&rez_v[rez++], &left[i]);
						copy = FALSE;
					}
				}
				FreeValues(left, amount_l);
				rez_v->type = INTEGER_T;
				rez_v->value.int_value = rez;
				left_array->array[proc_num].values = rez_v;
				left_array->array[proc_num].amount = (rez ? rez : 1);
				break;

			case NOT:
				if(left->type == INTEGER_T)
					left->value.int_value = !left->value.int_value;
				else
				{
					*error_message = BOOL_OP_ERROR;
					FreeValuesProcArray(left_array);
					FreeValuesProcArray(right_array);
					return NULL;
				}
				break;

			case AND:
				if((left->type == INTEGER_T) && (right->type == INTEGER_T))
					left->value.int_value = left->value.int_value && right->value.int_value;
				else
				{
					*error_message = BOOL_OP_ERROR;
					FreeValuesProcArray(left_array);
					FreeValuesProcArray(right_array);
					return NULL;
				}
				break;

			case OR:
				if((left->type == INTEGER_T) && (right->type == INTEGER_T))
					left->value.int_value = left->value.int_value || right->value.int_value;
				else
				{
					*error_message = BOOL_OP_ERROR;
					FreeValuesProcArray(left_array);
					FreeValuesProcArray(right_array);
					return NULL;
				}
				break;
			default:
				assert(FALSE);
			}
		} /* for(proc_num... */
		FreeValuesProcArray(right_array);
		return left_array; /* End of OperatorItem: */

	case ValueItem:
		if((left_array = MakeValuesProcArray(1, 1)) == NULL)
		{
			*error_message = OUT_MEMORY;
			return NULL;
		}
		memcpy(left_array->array->values, formula->item.value, sizeof(Value));
		assert(formula->item.value->indexes == NULL);
		return left_array;

	case FunctionItem:
		if(formula->item.function->type == WHEREFunction)
			return CalculateVariable(formula->item.function->operand->item.variable, error_message, TRUE);

		if((left_array = CalculateFormula(formula->item.function->operand, error_message)) == NULL)
			return NULL;

		for(proc_num = 0; proc_num < left_array->proc_amount; proc_num++)
		{
			Value *left, *rez_v;
			int i, rez, amount_l;

			left = left_array->array[proc_num].values;
			amount_l = left_array->array[proc_num].amount;
		
		
			if((rez_v = MakeValues((amount_l > MAX_DISPLAYED_VALUES ? MAX_DISPLAYED_VALUES : amount_l))) == NULL)
			{
				FreeValuesProcArray(left_array);
				*error_message = OUT_MEMORY;
				return NULL;
			}
			CopyRIValue(&rez_v[0], &left[0]);
			rez = 1;
			change = copy = FALSE;
			for(i = 1; i < amount_l; i++)
			{
				switch(rez_v[0].type)
				{
				case INTEGER_T:
					if(rez_v[0].value.int_value > left[i].value.int_value)
						change = TRUE;
					else if(rez_v[0].value.int_value == left[i].value.int_value)
						copy = TRUE;
					break;
				case REAL_T:
					if(rez_v[0].value.real_value > left[i].value.real_value)
						change = TRUE;
					else if(rez_v[0].value.real_value == left[i].value.real_value)
						copy = TRUE;
					break;
				case DOUBLE_T:
					if(rez_v[0].value.double_value > left[i].value.double_value)
						change = TRUE;
					else if(rez_v[0].value.double_value == left[i].value.double_value)
						copy = TRUE;
					break;
				default:
					assert(FALSE);
				}
				if((formula->item.function->type == MAXFunction) && !copy)
					change = !change;
				if(copy)
				{
					if(rez < MAX_DISPLAYED_VALUES)
						CopyRIValue(&rez_v[rez++], &left[i]);
					copy = FALSE;
				}
				else if(change)
				{
					CopyRIValue(&rez_v[0], &left[i]);
					while(rez > 1)
					{
						rez--;
						if(rez_v[rez].indexes)
						{
							free(rez_v[rez].indexes);
							rez_v[rez].indexes = NULL;
						}
					}
					change = FALSE;
				}
			}
			FreeValues(left, amount_l);
			left_array->array[proc_num].values = rez_v;
			left_array->array[proc_num].amount = rez;
		} /* for(proc_num ... */
		return left_array;
	}
	assert(FALSE);
	return NULL;
}

char *PrintValuesProcArray(const ValuesProcArray *value_array, int ID /* just for print */)
{
	char *buf;
	int i, j, pos;
	int total_number_of_values;
	int buffer_str_item_len = 0;

	for(j = 0, total_number_of_values = 0; j < value_array->proc_amount; j++)
	{
		if(value_array->array[j].amount)
		{
			if(value_array->array[j].values[0].type == CHARACTER_T)
			{
				for(i = 0; i < value_array->array[j].amount; i++)
					buffer_str_item_len += strlen(value_array->array[j].values[i].value.char_value);
			}
			else
				total_number_of_values += value_array->array[j].amount;
		}
	}


	if((buf = malloc(MAX_NUMBER_LEN + (MAX_NUMBER_LEN + MAX_INDEX_INFO_LEN) * 
		total_number_of_values + buffer_str_item_len +
		value_array->proc_amount + 1)) == NULL)
		return NULL;

	if(ID != -1)
	{
		sprintf(buf, "%d ", ID);
		pos = strlen(buf);
	}
	else
		pos = 0;

	for(j = 0, total_number_of_values = 0; j < value_array->proc_amount; j++)
	{
		Value *value = value_array->array[j].values;

		if(value_array->proc_amount > 1)
			buf[pos++] = '|';

		for(i = 0; i < value_array->array[j].amount; i++)
		{
			if(i)
				buf[pos++] = ' ';
			switch(value[i].type)
			{
			case INTEGER_T:
				sprintf(&buf[pos], "%d", value[i].value.int_value);
				break;
			case REAL_T:
				sprintf(&buf[pos], "%g", value[i].value.real_value);
				break;
			case DOUBLE_T:
				sprintf(&buf[pos], "%g", value[i].value.double_value);
				break;
			case CHARACTER_T:
				strcpy(&buf[pos], value[i].value.char_value);
				break;
			default:
				assert(FALSE);
			}
			pos += strlen(&buf[pos]);
			if(value[i].indexes)
			{
				buf[pos++] = '(';
				strncpy(&buf[pos], value[i].indexes, MAX_INDEX_INFO_LEN);
				pos += strlen(&buf[pos]);
				buf[pos++] = ')';
				buf[pos] = '\0';
			}
		}
	}
	buf[pos] = '\0';
	return buf;
}

/*--------------------------------------------------------*/
/* This function checks is the 2 values equal
   return value:
   0 - values equal
   1 - values not equal
   2 - values not comparable
*/
int ValueEqual(Value *v1, Value *v2)
{
	if(v1->type != v2->type)
		return 2; /* not comparable */
	switch(v1->type)
	{
	case INTEGER_T:
		return v1->value.int_value == v2->value.int_value ? 0 : 1;
	case REAL_T:
		return v1->value.real_value == v2->value.real_value ? 0 : 1;
	}
/*	case DOUBLE_T:*/
	assert(v1->type == DOUBLE_T);
	return v1->value.double_value == v2->value.double_value ? 0 : 1;
}
