#include <pthread.h>
#include <semaphore.h>
#include <stdio.h>
#include <malloc.h>
#include <string.h>
#include <ctype.h>
#include <syslog.h>

#include "TCPIPlib.h"
#include "CommonMonitorDefs.h"
#include "Thread_Semaphore.h"

#define MYNAME "gated"

typedef struct LoopPars
{
	TCPIPConnectionHandle handle[2]; /* 0 - UI, 1 - Monitor */
	BOOL *has_disconnected[2];
	sem_t finish_sem;
	sem_t *disconnect_sem;
	int id; /* 0 - UI, 1 - Monitor */
} LoopPars;

static pthread_attr_t attr; /* attributes for thread - for set them PTHREAD_CREATE_DETACHED */


static void PrintErrorMessage(const char *msg)
{
#ifdef DEBUG
	fputs(msg, stderr);
	fputc('\n',stderr);
#endif
	openlog(MYNAME, LOG_PID|LOG_CONS, LOG_DAEMON);
	syslog(LOG_ERR, msg);
	closelog();
}

static void loop_exit(LoopPars *pars)
{
	int i;

	sem_wait_not_intr(pars->disconnect_sem);
	for(i = 0; i < 2; i++)
		if(!*(pars->has_disconnected[i]))
		{
			TCPIPDisconnectEx(pars->handle[i], FALSE);
			*(pars->has_disconnected[i]) = TRUE;
		}
	sem_post(pars->disconnect_sem);
	if(pars->id == 0)
		sem_post(&(pars->finish_sem));
}

static THREADRETURN main_loop(void *arg)
{
	LoopPars *pars = (LoopPars *) arg;
	int mes_id;
	char *data;

#ifdef DEBUG
	fprintf(stderr, "main_loop: id: %d\n", pars->id);
#endif

	while(1)
	{
		/* get the message */
#ifdef DEBUG
		fprintf(stderr, "Before get message from %d\n", pars->id);
#endif
		if(!TCPIPGetMessage(pars->handle[pars->id], &mes_id, &data))
		{
#ifdef DEBUG
			fprintf(stderr, "Error get message from %d\n", pars->id);
#endif
			loop_exit(pars);
			return;
		}

#ifdef DEBUG
		fprintf(stderr, "Message from %s to %s: id: %d, data: %s\n",
			(pars->id == 0 ? "UI" : "Monitor"),
			(pars->id == 0 ? "Monitor" : "UI"),
			mes_id, (data ? data : "NULL"));
#endif

		/* send the message */
		if(!TCPIPSendMessage(pars->handle[1 - pars->id], mes_id, data))
		{
			PrintErrorMessage("Error send message");
			loop_exit(pars);
			return;
		}
		if(data)
		{
			free(data);
			data = NULL;
		}
	}
	return;
}


static THREADRETURN StartNewUI(void *arg)
{
	IPInfo mi;
	pthread_t thr;
	TCPIPConnectionHandle ui_handle = (TCPIPConnectionHandle) arg, monitor_handle;
	char buffer[MAX_NUMBER_LEN];
	LoopPars ui_pars, monitor_pars;
	sem_t disconnect_sem;
	BOOL ui_disconnect = FALSE, monitor_disconnect = FALSE;

	mi.type = ((int)ui_handle) + 3; /* wait Monitor with special ID */
	sprintf(buffer, "%d", mi.type);
	TCPIPSendMessage(ui_handle, SYS_GATE_ID, buffer);

#ifdef DEBUG
	puts("UI successfully connected");
	puts("Wait for the Monitor connect...");
#endif
	
	mi.is_server = TRUE;
	
	if(TCPIPGetConnectionFromThread(&mi, &monitor_handle, 100) == FALSE)
	{
		openlog(MYNAME, LOG_PID|LOG_CONS, LOG_DAEMON);
		syslog(LOG_INFO, "Monitor with ticket ID %d failed to connect by timeout", mi.type);
		closelog();
		TCPIPDisconnectEx(ui_handle, FALSE);
		return;
	}

	openlog(MYNAME, LOG_PID|LOG_CONS, LOG_DAEMON);
	syslog(LOG_INFO, "Monitor with ticket ID %d connected from %s", mi.type, mi.address);
	closelog();

#ifdef DEBUG
	puts("Monitor successfully connected");
#endif

	if(sem_init(&ui_pars.finish_sem, 0, 0))
	{
		PrintErrorMessage("error_init sem");
		return;
	}

	ui_pars.handle[0] = monitor_pars.handle[0] = ui_handle;
	ui_pars.handle[1] = monitor_pars.handle[1] = monitor_handle;
	ui_pars.id = 0;
	monitor_pars.id = 1;

	if(sem_init(&disconnect_sem, 0, 1))
	{
		PrintErrorMessage("error_init sem");
		return;
	}
	ui_pars.disconnect_sem = monitor_pars.disconnect_sem = &disconnect_sem;

	ui_pars.has_disconnected[0] = monitor_pars.has_disconnected[0] = &ui_disconnect;
	ui_pars.has_disconnected[1] = monitor_pars.has_disconnected[1] = &monitor_disconnect;

	if(pthread_create(&thr, &attr, main_loop, (void*)&ui_pars))
	{
		PrintErrorMessage("error_init thread");
		return;
	}

	main_loop((void*)&monitor_pars);

	sem_wait_not_intr(&ui_pars.finish_sem);
	sem_destroy(&ui_pars.finish_sem);
	sem_destroy(&disconnect_sem);

	openlog(MYNAME, LOG_PID|LOG_CONS, LOG_DAEMON);
	syslog(LOG_INFO, "Session with ticket ID %d has been finished successfully", mi.type);
	closelog();

	return;
}

int main(int argc, char **argv)
{
	int i;
	IPInfo mi;
	pthread_t thr;
	TCPIPConnectionHandle tcp_ip_handle;
	char *error = NULL;
	int fd;
	struct rlimit flim;

	puts("Gate for Norma debugger or ComDeb started");
	if((argc > 1) && !isdigit(argv[1][0]))
	{
		puts("Usage: gate [port]");
		return 1;
	}

	/* Init the attributes for thread creating */
	pthread_attr_init(&attr);
	pthread_attr_setdetachstate(&attr, PTHREAD_CREATE_DETACHED);
	pthread_attr_setstacksize(&attr, 2 * PTHREAD_STACK_MIN); /* We don't need the large stack !!! */

	
	mi.is_server = TRUE;
	
	if(argc > 1)
		mi.port_number = atoi(argv[1]);
	else
		mi.port_number = 22;

	mi.type = 0; /* wait UI */

	error_message_pointer = &error;
	TCPIPSetMaxPortsNum(1);
	if(!TCPIPServerCreateSocket(&mi))
	{
		openlog(MYNAME, LOG_PID|LOG_CONS, LOG_DAEMON);
		syslog(LOG_ERR, "Failed to create server on port %d", mi.port_number);
		fprintf(stderr, "Failed to create server on port %d\n", mi.port_number);
		if(error)
		{
			syslog(LOG_ERR, error);
			fprintf(stderr, "%s\n", error);
		}
		closelog();
		return 1;
	}
#ifndef DEBUG
/*    init -   ,
  -    :
SIGTTOU, SIGTTIN, SIGTSTP*/
	if(getppid() != 1)
	{
		signal(SIGTTOU, SIG_IGN);
		signal(SIGTTIN, SIG_IGN);
		signal(SIGTSTP, SIG_IGN);
/*      
( shell  ..)*/
		if(fork() != 0)
			exit(0);
/*     ,   
  */
		setsid();
	}
/*   */
	getrlimit(RLIMIT_NOFILE, &flim);
	for(fd = 0; fd < 3 /*flim.rlim_max*/; fd++)
		close(fd);
/*    */
	chdir("/");
#endif

/*   */
	openlog(MYNAME, LOG_PID|LOG_CONS, LOG_DAEMON);
	syslog(LOG_INFO, "Gate for Norma debugger or ComDeb started on port %d", mi.port_number);
	closelog();

	while(1)
	{
#ifdef DEBUG
		puts("Wait for the UI connect...");
#endif
		if(TCPIPConnect(&mi, &tcp_ip_handle, NULL) == FALSE)
		{
			if(error)
				PrintErrorMessage(error);
			continue;
		}

		openlog(MYNAME, LOG_PID|LOG_CONS, LOG_DAEMON);
		syslog(LOG_INFO, "UI %s connected", mi.address);
		closelog();

		if(pthread_create(&thr, &attr, StartNewUI, (void*)tcp_ip_handle))
		{
			PrintErrorMessage("error_init thread, Gate exiting");
			return 1;
		}
	}

	return 1;
}

