/* Monitor-communication library */
/* All these functions called from debugging program code */

#include <stdlib.h>
#include <stdio.h>
#include <stdarg.h>

#include "mpi.h"
#include "MonitorConnect.h"
#include "TCPIPlib.h"
#include "NDMessages.h"

/*#define DEBUG*/
#ifdef WIN32
#define FORTRAN_PRINT_STR "%d "
#define FORTRAN_PRINT_STR5 "%d %d %d %d %d"
#define FORTRAN_PRINT_STR6 "%d %d %d %d %d %d"
#else
#define FORTRAN_PRINT_STR "%d "
#define FORTRAN_PRINT_STR5 "%d %d %d %d %d"
#define FORTRAN_PRINT_STR6 "%d %d %d %d %d %d"
#endif

static BOOL is_monitor_connect_inited = FALSE;
static TCPIPConnectionHandle tcp_ip_handle;
static FORTRAN_INTEGER type_int_size, type_real_size, type_double_size;
static BOOL was_error = FALSE;

#ifdef DEBUG
static char *glodal_debug_pointer;
#endif

static VarInfo *global_var_info = NULL;

static void PrintErrorMessage(const char *msg) {fputs(msg, stderr);fputc('\n',stderr);}
static void out_memory(void) {PrintErrorMessage(OUT_MEMORY);}

static const char *ConvertString2C(const char *s)
{
	int i;
	static char buf[MAX_FILE_NAME+1];
	for(i = 0; s[i]; i++)
		if(memcmp(&s[i], "\\000", 4) == 0)
		{
			assert(i < MAX_FILE_NAME);
			memcpy(buf, s, i);
			buf[i] = '\0';
			return buf;
		}
	return s;
}	

static BOOL ReadMonitorInfoFile(const char *MonitorInfoFileName,
								IPInfo *mi)
{
	FILE *f;
	char buffer[MAX_FILE_NAME];

	if((f = fopen(MonitorInfoFileName, "r")) == NULL)
	{
		sprintf(buffer, MONITOR_INFO_FILE_NOTFOUND, MonitorInfoFileName);
		PrintErrorMessage(buffer);
		return FALSE;
	}
	
	if(fscanf(f, "%s\n", mi->address) != 1)
		goto monitor_format_error;
	if(fscanf(f, "%d\n", &mi->port_number) != 1)
		goto monitor_format_error;
	if(fgets(buffer, sizeof(buffer), f) == NULL)
		goto monitor_format_error;
	if(memcmp(buffer, SERVER_ID, sizeof(SERVER_ID)) == 0) /* i.e. Monitor is Server */
		mi->is_server = FALSE;
	else if(memcmp(buffer, CLIENT_ID, sizeof(CLIENT_ID)) == 0) /* i.e. Monitor is Client */
		mi->is_server = TRUE;
	else 
		goto monitor_format_error;

	fclose(f);
	mi->type = 1; /* we are the debugged program */
	return TRUE;

monitor_format_error:
	fclose(f);
	sprintf(buffer, MONITOR_INFO_FILE_FORMATERROR, MonitorInfoFileName);
	PrintErrorMessage(buffer);
	return FALSE;
}


/* Initializing the connect with Monitor
   Send to Monitor my coordinates */
static BOOL InitMonitorConnect(void /*int myi, int myj*/)
{
	IPInfo mi;

	if(is_monitor_connect_inited)
		return TRUE;
	if(ReadMonitorInfoFile(MONITOR_INFO_FILENAME, &mi) == FALSE)
		return FALSE;
	assert(mi.is_server == FALSE);
	if(TCPIPConnect(&mi, &tcp_ip_handle, NULL) == FALSE)
	{
/*		sprintf(buffer, "%d %d: Failed connect to monitor\n", myi, myj);
		PrintErrorMessage(buffer);*/
		PrintErrorMessage("Failed connect to monitor\n");
		return FALSE;
	}
	is_monitor_connect_inited = TRUE;
/*	sprintf(buffer, "%d %d", myi, myj);
	TCPIPSendMessage(tcp_ip_handle, ND_NOTIF_MY_IDS, buffer);*/
	return TRUE;
}

static void CloseMonitorConnect(void)
{
	TCPIPDisconnect(tcp_ip_handle);
	is_monitor_connect_inited = FALSE;
}

/* *************************************************** */
/* Part INFO routines */
/* *************************************************** */
typedef struct PartInfo
{
	char *name;
	int i, j, myi, myj, ipoints, jpoints;
	char iname[MAX_VARIABLE_NAME], jname[MAX_VARIABLE_NAME];
	BOOL is_C, is_External;
	VarInfo *vars;
	struct PartInfo *next;
} PartInfo;

static PartInfo *cur_part_info = NULL;

static void StartNewPart(const char *name, BOOL is_C, BOOL is_External)
{
	PartInfo *new_part_info;

	if((new_part_info = (PartInfo *)malloc(sizeof(PartInfo))) == NULL)
	{
		out_memory();
		return;
	}
	new_part_info->next = cur_part_info;
	cur_part_info = new_part_info;
	new_part_info->vars = NULL;
	if((new_part_info->name = (char *)malloc(strlen(name) + 1)) == NULL)
	{
		out_memory();
		return;
	}
	strcpy(new_part_info->name, name);
	new_part_info->is_C = is_C;
	new_part_info->is_External = is_External;
}

static void SetPartDistrParameters(int i, int j, int myi, int myj,
								   int ipoints, int jpoints,
								   const char *iname, const char *jname)
{
	assert(cur_part_info);
	cur_part_info->i = i;
	cur_part_info->j = j;
	cur_part_info->myi = myi;
	cur_part_info->myj = myj;
	cur_part_info->ipoints = ipoints;
	cur_part_info->jpoints = jpoints;
	strcpy(cur_part_info->iname, iname);
	strcpy(cur_part_info->jname, jname);
}

static void SetPartDistrParametersNULL(void)
{
	SetPartDistrParameters(0, 0, 0, 0, 0, 0, "", "");
}

static BOOL IsDisributedPart(void)
{
	assert(cur_part_info);
	return(cur_part_info->i != 0);
}

static void FreeVarInfo(VarInfo *vars)
{
	VarInfo *tek;
	while(vars)
	{
		tek = vars;
		vars = vars->next;
		if(tek->number_of_dimensions)
			free(tek->dimensions);
		free(tek);
	}
}

static void EndPart(void)
{
	PartInfo *was_part_info = cur_part_info;
	cur_part_info = cur_part_info->next;
	free(was_part_info->name);
	FreeVarInfo(was_part_info->vars);
	free(was_part_info);
}

static void NewVarInfo(const char *name, void *pointer, NormaVarType type,
					   int number_of_dimensions, DimInfo *dimensions,
					   BOOL is_global)
{
	VarInfo *new_var_info;

	if((new_var_info = (VarInfo *)malloc(sizeof(VarInfo))) == NULL)
	{
		out_memory();
		return;
	}

	assert(strlen(name) < MAX_VARIABLE_NAME);
	strcpy(new_var_info->name, name);

	if(type >= CHARACTER_ID_BASE)
	{
		new_var_info->type = CHARACTER_T;
		new_var_info->len = type - CHARACTER_ID_BASE;
	}
	else
		new_var_info->type = type;
	new_var_info->pointer = pointer;
	new_var_info->number_of_dimensions = number_of_dimensions;
	new_var_info->dimensions = dimensions;
	if(cur_part_info && !is_global)
	{
		new_var_info->next = cur_part_info->vars;
		cur_part_info->vars = new_var_info;
	}
	else
	{
		new_var_info->next = global_var_info;
		global_var_info = new_var_info;
	}
}

/* *************************************************** */
static int EncodeStr(char *buffer, const char *str, int maxlen /* -1 if not bounded */)
{
	int buf_pos = 0;
	int str_pos = 0;

	while((str_pos != maxlen) && str[str_pos])
	{
		if((str[str_pos] == '\"') || (str[str_pos] == '\\'))
			buffer[buf_pos++] = '\\';
		buffer[buf_pos++] = str[str_pos++];
	}
	return buf_pos;
}

static int GetVariable(char *value, VarInfo *vars, int dim,
						int *previous_dim_value, BOOL retrieve_index_info,
						BOOL is_C)
{
	int i, j;
	int base_smes, pos = 0;
	int block;

	if(dim == 0)
	{ /* the last dim - calculate the value */

		base_smes = 0;

		block = 1;
		previous_dim_value[0] = vars->dimensions[0].requested_start;
		if(is_C)
		{
			for(i = vars->number_of_dimensions - 1; ; i--)
			{
				base_smes += (previous_dim_value[i] - vars->dimensions[i].start) * block;
				if(i > 0)
					block *= (vars->dimensions[i].end - vars->dimensions[i].start + 1);
				else
					break;
			}
		}
		else /* Fortran */
		{
			for(i = 0; i < vars->number_of_dimensions; i++)
			{
				base_smes += (previous_dim_value[i] - vars->dimensions[i].start) * block;
				block *= (vars->dimensions[i].end - vars->dimensions[i].start + 1);
			}
		}
	}
	
	for(i = vars->dimensions[dim].requested_start; i <= vars->dimensions[dim].requested_end; i++)
	{
		if(dim == 0)
		{ /* the last dim - calculate the value */
			switch(vars->type)
			{
			case INTEGER_T:
				if(is_C)
					sprintf(&value[pos], "%d ", ((int *)vars->pointer)[base_smes]);
				else
					sprintf(&value[pos], FORTRAN_PRINT_STR, ((FORTRAN_INTEGER *)vars->pointer)[base_smes]);
				break;
			case REAL_T:
				sprintf(&value[pos], "%g ", ((float *)vars->pointer)[base_smes]);
				break;
			case DOUBLE_T:
				sprintf(&value[pos], "%g ", ((double *)vars->pointer)[base_smes]);
				break;
			case CHARACTER_T:
				value[pos++] = ' ';
				value[pos++] = '\"';
				pos += EncodeStr(&value[pos], &(((char *)vars->pointer)[base_smes * vars->len]), vars->len);
				value[pos++] = '\"';
				value[pos] = 0;
				break;
			default:
				assert(FALSE);
			}
			if(is_C)
				base_smes += block;
			else
				base_smes++;
			pos += strlen(&value[pos]);
			if(retrieve_index_info)
			{
#ifdef DEBUG
/*				fprintf(stderr, "GetVariable: smes: %d\n", 
					value - glodal_debug_pointer + pos);*/
#endif
				value[pos++] = '(';
				for(j = 0; j < vars->number_of_dimensions; j++)
				{
					int smes;
					if(j > 0)
					{
						value[pos++] = ',';
						value[pos++] = ' ';
					}
					if(strcmp(vars->dimensions[j].name, cur_part_info->iname) == 0)
						smes = (cur_part_info->myi - 1) * cur_part_info->ipoints;
					else if(strcmp(vars->dimensions[j].name, cur_part_info->jname) == 0)
						smes = (cur_part_info->myj - 1) * cur_part_info->jpoints;
					else
						smes = 0;
					sprintf(&value[pos], "%s=%d", 
						vars->dimensions[j].name,
						smes + (j == 0 ? i : previous_dim_value[j]));
					pos += strlen(&value[pos]);
				}
				value[pos++] = ')';
				value[pos] = '\0';
			}
		}
		else /* recursive call */
		{
			previous_dim_value[dim] = i;
			pos += GetVariable(&value[pos], vars, dim - 1, 
				previous_dim_value, retrieve_index_info, is_C);
		}
	}
	return pos;
}

static int GetMonitorAnswer(void)
{
	char *data;
	TCPIPMessageType rez;

wait:
#ifdef DEBUG
	fprintf(stderr, "GetMonitorAnswer: before get the message\n");
#endif
	if(!TCPIPGetMessage(tcp_ip_handle, &rez, &data))
	{
		PrintErrorMessage("Error get message from Monitor");
		CloseMonitorConnect();
		MPI_Finalize();
		exit(1);
	}

#ifdef DEBUG
	fprintf(stderr, "GetMonitorAnswer: after get the message (%d)\n", rez);
#endif
	if(rez == ND_CONTINUE_EXECUTION)
	{
#ifdef DEBUG
		fprintf(stderr, "GetMonitorAnswer: ND_CONTINUE_EXECUTION\n");
#endif
		return 1;
	}
	if(rez == ND_GET_VALUE)
	{ /* get the variable value */
		int pos, i, amount, item_len;
		VarInfo *vars;
		char err_msg[] = "Internal error", *value;
		BOOL is_Scalar;
		BOOL retrieve_index_info;

#ifdef DEBUG
		fprintf(stderr, "GetMonitorAnswer: ND_GET_VALUE: %s\n", data);
#endif
		assert(data);
		pos = 0;
		while((data[pos] != '\0') && (data[pos] != ' '))
			pos++;
		if(data[pos] == '\0')
			is_Scalar = TRUE;
		else
		{
			is_Scalar = FALSE;
			data[pos++] = 0;
			retrieve_index_info = (data[pos++] != '0');
		}
		assert(cur_part_info);
		vars = cur_part_info->vars;
		while(vars)
		{
			if(strcmp(vars->name, data) == 0)
				break;
			vars = vars->next;
		}
		if(vars == NULL)
		{ /* search in global variables */
			vars = global_var_info;
			while(vars)
			{
				if(strcmp(vars->name, data) == 0)
					break;
				vars = vars->next;
			}
		}
		if((vars == NULL) ||
			(is_Scalar != (vars->number_of_dimensions == 0)))
		{
		error_get_value:
			TCPIPSendMessage(tcp_ip_handle, ND_TAKE_VALUE_ERROR, err_msg);
			goto wait;
		}
		pos++;
		amount = 1;
		for(i = 0; i < vars->number_of_dimensions; i++)
		{
			assert(!is_Scalar);
			while(data[pos] == ' ')
				pos++;
			if(sscanf(&data[pos], "%d %d",
				&vars->dimensions[i].requested_start, 
				&vars->dimensions[i].requested_end) != 2)
				goto error_get_value;
			while(isdigit(data[pos]))
				pos++;
			while(data[pos] == ' ')
				pos++;
			while(isdigit(data[pos]))
				pos++;
			amount = amount * (vars->dimensions[i].requested_end - vars->dimensions[i].requested_start + 1);
		}

#ifdef DEBUG
		fprintf(stderr, "GetMonitorAnswer: amount: %d\n", amount);
#endif
		if(vars->type == CHARACTER_T)
			item_len = vars->len + 4 /* for space, quotas and NULL symbol */;
		else
			item_len = MAX_NUMBER_LEN + (retrieve_index_info ? MAX_INDEX_INFO_LEN : 0);
		if((value = (char *)malloc(amount * item_len)) == NULL)
		{
			out_memory();
			TCPIPSendMessage(tcp_ip_handle, ND_TAKE_VALUE_ERROR, err_msg);
			goto wait;
		}
#ifdef DEBUG
		glodal_debug_pointer = value;
#endif
		
		if(is_Scalar)
		{ /* Scalar */
			switch(vars->type)
			{
			case INTEGER_T:
				if(cur_part_info->is_C)
					sprintf(value, "%d", *((int *)vars->pointer));
				else
					sprintf(value, FORTRAN_PRINT_STR, *((FORTRAN_INTEGER *)vars->pointer));
				break;
			case REAL_T:
				sprintf(value, "%g", *((float *)vars->pointer));
				break;
			case DOUBLE_T:
				sprintf(value, "%g", *((double *)vars->pointer));
				break;
			case CHARACTER_T:
				value[0] = '\"';
				pos = 1;
				pos += EncodeStr(&value[1], (char *)vars->pointer, vars->len);
				value[pos++] = '\"';
				value[pos++] = 0;
				break;
			default:
				assert(FALSE);
			}
		}
		else
		{ /* VARIABLE ON DOMAIN */
			int *previous_dim_value;

			if((previous_dim_value = (int *)malloc(sizeof(int) * vars->number_of_dimensions)) == NULL)
			{
				free(value);
				out_memory();
				TCPIPSendMessage(tcp_ip_handle, ND_TAKE_VALUE_ERROR, err_msg);
				goto wait;
			}

			pos = GetVariable(value, vars, vars->number_of_dimensions - 1,
				previous_dim_value, retrieve_index_info, cur_part_info->is_C);
			free(previous_dim_value);
		}
		/* send the retrieved value */
#ifdef DEBUG
		fprintf(stderr, "GetMonitorAnswer: send value: %s\n", value);
#endif
		TCPIPSendMessage(tcp_ip_handle, ND_TAKE_VALUE, value);
		free(value);
		goto wait;
	}
	if(rez == ND_BREAK_EXECUTION)
	{/* Break signal. Stop. */
		TCPIPSendMessage(tcp_ip_handle, ND_BREAK_EXECUTION, NULL);
#ifdef DEBUG
		fprintf(stderr, "GetMonitorAnswer: after send ND_BREAK_EXECUTION\n");
#endif
		CloseMonitorConnect();
		MPI_Finalize();
		exit(1);
	}
	if(rez == ND_GET_VERSION_ID)
	{
		char buf[100];
		sprintf(buf, 
			"MonitorConnect library: version %s built at %s.",
			VERSIONID, __DATE__);
		TCPIPSendMessage(tcp_ip_handle, ND_GET_VERSION_ID, buf);
		goto wait;
	}
/*
	if(cur_part_info)
		fprintf(stderr, "rez = %d myi = %d myj = %d\n", rez, cur_part_info->myi, cur_part_info->myj);
	else
		fprintf(stderr, "rez = %d\n", rez);
*/
	assert(FALSE);
	return 0;
}
/* *************************************************** */
/* Public functions */
/* *************************************************** */
/* Start of all program
   This function should initialize the Monitor */ 
int NDStartProgram(const char *main_part_name, const char *filename,
			 FORTRAN_INTEGER start_line, FORTRAN_INTEGER CRC, FORTRAN_INTEGER number_of_tasks, FORTRAN_INTEGER mynumber)
{
	return FNAME(ndstprg)(main_part_name, filename,
			 &start_line, &CRC, &number_of_tasks, &mynumber);
}

int FNAME(ndstprg)(const char *main_part_name, const char *filename, 
			 FORTRAN_INTEGER *start_line, FORTRAN_INTEGER *CRC, FORTRAN_INTEGER *number_of_tasks, FORTRAN_INTEGER *mynumber)
{
	char buffer[MAX_FILE_NAME+MAX_PART_NAME+MAX_NUMBER_LEN];

	assert(strlen(ConvertString2C(main_part_name)) < MAX_PART_NAME);
	assert(strlen(ConvertString2C(filename)) < MAX_FILE_NAME);

	if(!InitMonitorConnect(/*0, 0*/))
	{
		was_error = TRUE;
		return 0; /* Initialization fails. Stop */
	}
	assert(is_monitor_connect_inited);
	sprintf(buffer, FORTRAN_PRINT_STR, *number_of_tasks);
	sprintf(&buffer[strlen(buffer)], FORTRAN_PRINT_STR, *mynumber);
	TCPIPSendMessage(tcp_ip_handle, ND_TASK_ID, buffer);

	strcpy(buffer, ConvertString2C(main_part_name));	
	sprintf(&buffer[strlen(buffer)], " \"%s\" ", ConvertString2C(filename));
	sprintf(&buffer[strlen(buffer)], FORTRAN_PRINT_STR, *start_line);
	sprintf(&buffer[strlen(buffer)], FORTRAN_PRINT_STR, *CRC);
	
	TCPIPSendMessage(tcp_ip_handle, ND_NOTIF_PROGRAM_START, buffer);
	GetMonitorAnswer();
	return 1;
}
/* Fortran program */
int FNAME(ndstpex)(const char *program_name, const char *filename, FORTRAN_INTEGER *start_line, FORTRAN_INTEGER *CRC)
{
	char buffer[MAX_FILE_NAME+MAX_NUMBER_LEN];
	int size;

	assert(strlen(ConvertString2C(filename)) < MAX_FILE_NAME);

	if(!InitMonitorConnect(/*0, 0*/))
	{
		was_error = TRUE;
		return 0; /* Initialization fails. Stop */
	}
	assert(is_monitor_connect_inited);
	if(MPI_Comm_size(MPI_COMM_WORLD, &size) != MPI_SUCCESS)
	{
		PrintErrorMessage("Error call MPI_Comm_size");
		was_error = TRUE;
		return 0; /* Initialization fails. Stop */
	}

	sprintf(buffer, "%d -1", size);
	TCPIPSendMessage(tcp_ip_handle, ND_TASK_ID, buffer);

	strcpy(buffer, ConvertString2C(program_name));	
	sprintf(&buffer[strlen(buffer)], " \"%s\" ", ConvertString2C(filename));
	sprintf(&buffer[strlen(buffer)], FORTRAN_PRINT_STR, *start_line);
	sprintf(&buffer[strlen(buffer)], FORTRAN_PRINT_STR, *CRC);
	
	TCPIPSendMessage(tcp_ip_handle, ND_NOTIF_PROGRAM_START, buffer);
	StartNewPart(ConvertString2C(program_name), TRUE, TRUE);
	SetPartDistrParametersNULL();
	GetMonitorAnswer();
	return 1;
}

void NDEndProgram(const char *dummy, FORTRAN_INTEGER dummy1)
{
	FNAME(ndenprg)();
}

void FNAME(ndenprg)(void)
{
	if(was_error)
		return;
	assert(is_monitor_connect_inited);
	TCPIPSendMessage(tcp_ip_handle, ND_NOTIF_PROGRAM_END, NULL);
	FreeVarInfo(global_var_info);
	global_var_info = NULL;
	CloseMonitorConnect();
}

/*         Till now unused         */
void NDRegisterSize(FORTRAN_INTEGER int_size, FORTRAN_INTEGER real_size, FORTRAN_INTEGER double_size)
{
	FNAME(ndregsz)(&int_size, &real_size, &double_size);
}

void FNAME(ndregsz)(FORTRAN_INTEGER *int_size, FORTRAN_INTEGER *real_size, FORTRAN_INTEGER *double_size)
{
	type_int_size = *int_size;
	type_real_size = *real_size; 
	type_double_size = *double_size;
}

/*=================================================================*/
static void StartMainPart(const char *name, const char *filename, 
						  FORTRAN_INTEGER *start_line, FORTRAN_INTEGER *CRC, BOOL is_C)
{
	char buffer[MAX_FILE_NAME+MAX_PART_NAME+MAX_NUMBER_LEN];

	if(was_error)
		return;

	assert(is_monitor_connect_inited);
	assert(strlen(ConvertString2C(name)) < MAX_PART_NAME);
	assert(strlen(ConvertString2C(filename)) < MAX_FILE_NAME);

	strcpy(buffer, ConvertString2C(name));	
	sprintf(&buffer[strlen(buffer)], " \"%s\" ", ConvertString2C(filename));
	sprintf(&buffer[strlen(buffer)], FORTRAN_PRINT_STR, *start_line);
	sprintf(&buffer[strlen(buffer)], FORTRAN_PRINT_STR, *CRC);
	
	TCPIPSendMessage(tcp_ip_handle, ND_NOTIF_MAIN_PART_START, buffer);
	StartNewPart(ConvertString2C(name), is_C, FALSE);
	SetPartDistrParametersNULL();
	GetMonitorAnswer();
}


/* Start of Main Part */ 
void NDStartMainPart(const char *name, const char *filename, FORTRAN_INTEGER start_line, FORTRAN_INTEGER CRC)
{
	StartMainPart(name, filename, &start_line, &CRC, TRUE);
}

void FNAME(ndstmp)(const char *name, const char *filename, FORTRAN_INTEGER *start_line, FORTRAN_INTEGER *CRC)
{
	StartMainPart(name, filename, start_line, CRC, FALSE);
}

void NDEndMainPart(const char *filename, FORTRAN_INTEGER end_line)
{
	FNAME(ndenmp)(filename, &end_line);
}

void FNAME(ndenmp)(const char *filename, FORTRAN_INTEGER *end_line)
{
	char buffer[/*MAX_FILE_NAME + */MAX_NUMBER_LEN];

	if(was_error)
		return;

	assert(is_monitor_connect_inited);
	
//	sprintf(buffer, "\"%s\" %d", ConvertString2C(filename), *end_line);
	sprintf(buffer, FORTRAN_PRINT_STR, *end_line);
	TCPIPSendMessage(tcp_ip_handle, ND_NOTIF_MAIN_PART_END, buffer);
	GetMonitorAnswer();
	EndPart();
}

/*==============================================================*/
static int StartPart(const char *name, const char *filename, 
					 FORTRAN_INTEGER *start_line, FORTRAN_INTEGER *CRC, BOOL is_C, BOOL is_External)
{
	char buffer[MAX_FILE_NAME+MAX_PART_NAME+MAX_NUMBER_LEN];

	if(was_error)
		return 0;
	assert(is_monitor_connect_inited);
	assert(strlen(ConvertString2C(name)) < MAX_PART_NAME);
	assert(strlen(ConvertString2C(filename)) < MAX_FILE_NAME);

	strcpy(buffer, ConvertString2C(name));	
	sprintf(&buffer[strlen(buffer)], " \"%s\" ", ConvertString2C(filename));
	sprintf(&buffer[strlen(buffer)], FORTRAN_PRINT_STR, *start_line);
	sprintf(&buffer[strlen(buffer)], FORTRAN_PRINT_STR, *CRC);
	sprintf(&buffer[strlen(buffer)], "%d", is_External);

	TCPIPSendMessage(tcp_ip_handle, ND_NOTIF_PART_START, buffer);
	StartNewPart(ConvertString2C(name), is_C, is_External);
	SetPartDistrParametersNULL();
	GetMonitorAnswer();
	return 1; /* All OK */
}
/* Scalar part */
int NDStartPart(const char *name, const char *filename, FORTRAN_INTEGER start_line, FORTRAN_INTEGER CRC)
{
	return StartPart(name, filename, &start_line, &CRC, TRUE, FALSE);
}

int FNAME(ndstp)(const char *name, const char *filename, FORTRAN_INTEGER *start_line, FORTRAN_INTEGER *CRC)
{
	return StartPart(name, filename, start_line, CRC, FALSE, FALSE);
}

void NDEndPart(const char *filename, FORTRAN_INTEGER end_line)
{
	FNAME(ndenp)(filename, &end_line);
}
void FNAME(ndenp)(const char *filename, FORTRAN_INTEGER *end_line)
{
	char buffer[/*MAX_FILE_NAME + */ MAX_NUMBER_LEN];

	if(was_error)
		return;

	assert(is_monitor_connect_inited);

//	sprintf(buffer, "\"%s\" %d", ConvertString2C(filename), *end_line);
	sprintf(buffer, FORTRAN_PRINT_STR, *end_line);
	TCPIPSendMessage(tcp_ip_handle, ND_NOTIF_PART_END, buffer);
	GetMonitorAnswer();
	EndPart();
}

/* External Fortran function */
int FNAME(ndstexf)(const char *name, const char *filename, FORTRAN_INTEGER *start_line, FORTRAN_INTEGER *CRC)
{
	return StartPart(name, filename, start_line, CRC, FALSE, TRUE);
}
void FNAME(ndenexf)(const char *filename, FORTRAN_INTEGER *end_line)
{
	FNAME(ndenp)(filename, end_line);
}

/*=============================================================*/
static void StartDistributionPart(const char *name, const char *filename, FORTRAN_INTEGER *start_line, FORTRAN_INTEGER *CRC,
	FORTRAN_INTEGER *i, FORTRAN_INTEGER *j, FORTRAN_INTEGER *myi, FORTRAN_INTEGER *myj, FORTRAN_INTEGER *ipoints, FORTRAN_INTEGER *jpoints,
	const char *iname, const char *jname, BOOL is_C)
{
	char buffer[256 + MAX_VARIABLE_NAME*2 + MAX_FILE_NAME + MAX_PART_NAME];

	assert(is_monitor_connect_inited);
	assert(strlen(ConvertString2C(name)) < MAX_PART_NAME);
	assert(strlen(ConvertString2C(filename)) < MAX_FILE_NAME);
	assert(strlen(ConvertString2C(iname)) < MAX_VARIABLE_NAME);
	assert(strlen(ConvertString2C(jname)) < MAX_VARIABLE_NAME);

	strcpy(buffer, ConvertString2C(name));	
	sprintf(&buffer[strlen(buffer)], " \"%s\" ", ConvertString2C(filename));
	sprintf(&buffer[strlen(buffer)], FORTRAN_PRINT_STR, *start_line);
	sprintf(&buffer[strlen(buffer)], FORTRAN_PRINT_STR, *CRC);
	sprintf(&buffer[strlen(buffer)], FORTRAN_PRINT_STR, *i);
	sprintf(&buffer[strlen(buffer)], FORTRAN_PRINT_STR, *j);
	sprintf(&buffer[strlen(buffer)], FORTRAN_PRINT_STR, *myi);
	sprintf(&buffer[strlen(buffer)], FORTRAN_PRINT_STR, *myj);
	sprintf(&buffer[strlen(buffer)], FORTRAN_PRINT_STR, *ipoints);
	sprintf(&buffer[strlen(buffer)], FORTRAN_PRINT_STR, *jpoints);
	strcat(buffer, ConvertString2C(iname));
	strcat(buffer, " ");
	strcat(buffer, ConvertString2C(jname));
	
	TCPIPSendMessage(tcp_ip_handle, ND_NOTIF_DISTRIBUTION_PART_START, buffer);
	StartNewPart(ConvertString2C(name), is_C, FALSE);
	strcpy(buffer, ConvertString2C(iname));
	SetPartDistrParameters(*i, *j, *myi, *myj, *ipoints, *jpoints,
		buffer, ConvertString2C(jname));
	GetMonitorAnswer();
/*	fprintf(stderr, "Monitor Connect: Distr part myi = %d myj = %d\n", cur_part_info->myi, cur_part_info->myj);*/
}

/* Distribution part */
void NDStartDistributionPart(const char *name, const char *filename, FORTRAN_INTEGER start_line, FORTRAN_INTEGER CRC,
	FORTRAN_INTEGER i, FORTRAN_INTEGER j, FORTRAN_INTEGER myi, FORTRAN_INTEGER myj, FORTRAN_INTEGER ipoints, FORTRAN_INTEGER jpoints,
	const char *iname, const char *jname)
{
	StartDistributionPart(name, filename, &start_line, &CRC,
		&i, &j, &myi, &myj, &ipoints, &jpoints,
		iname, jname, TRUE);
}

void FNAME(ndstdp)(const char *name, const char *filename, FORTRAN_INTEGER *start_line, FORTRAN_INTEGER *CRC,
	FORTRAN_INTEGER *i, FORTRAN_INTEGER *j, FORTRAN_INTEGER *myi, FORTRAN_INTEGER *myj, FORTRAN_INTEGER *ipoints, FORTRAN_INTEGER *jpoints,
	const char *iname, const char *jname)
{
	StartDistributionPart(name, filename, start_line, CRC,
		i, j, myi, myj, ipoints, jpoints,
		iname, jname, FALSE);
}

void NDEndDistributionPart(const char *filename, FORTRAN_INTEGER end_line)
{
	FNAME(ndendp)(filename, &end_line);
}

void FNAME(ndendp)(const char *filename, FORTRAN_INTEGER *end_line)
{
	char buffer[/*MAX_FILE_NAME + */MAX_NUMBER_LEN];

	if(was_error)
		return;

	assert(is_monitor_connect_inited);
	
//	sprintf(buffer, "\"%s\" %d", ConvertString2C(filename), *end_line);
	sprintf(buffer, FORTRAN_PRINT_STR, *end_line);
	TCPIPSendMessage(tcp_ip_handle, ND_NOTIF_DISTRIBUTION_PART_END, buffer);
	GetMonitorAnswer();
	EndPart();
}

/*==========================================================*/

static void RegisterVariable(const char *name, void *pointer, 
			NormaVarType type, int number_of_dimensions, 
			BOOL is_C, BOOL is_External, BOOL is_global, va_list args)
{
	DimInfo *dimensions;
	char buffer[200 + MAX_VARIABLE_NAME];

	assert(!is_global || is_External);

	sprintf(buffer, "%s %d %d", ConvertString2C(name), type, number_of_dimensions);
#ifdef DEBUG
	fprintf(stderr, "ndregvar_: before reg var  %s\n", ConvertString2C(name));
#endif
	if(!is_External)
		TCPIPSendMessage(tcp_ip_handle, ND_REGISTER_VAR, buffer);

	if(number_of_dimensions > 0) /* not scalar */
	{
		int i;

		if((dimensions = (DimInfo *)malloc(sizeof(DimInfo) * number_of_dimensions)) == NULL)
		{
			out_memory();
			return;
		}

		for(i = 0; i < number_of_dimensions; i++)
		{
			if(is_External)
			{
				if(is_C)
				{
				}
				else
				{
					FORTRAN_INTEGER *value;

					value = va_arg(args, FORTRAN_INTEGER *);
					dimensions[i].start = *value;
					value = va_arg(args, FORTRAN_INTEGER *);
					dimensions[i].end = *value;
				}
				sprintf(&buffer[strlen(buffer)], " %d %d", dimensions[i].start, dimensions[i].end);
			}
			else /* Norma var */
			{
				char *index_name;

				index_name = va_arg(args, char *);
				assert(strlen(ConvertString2C(index_name)) < MAX_VARIABLE_NAME);
				strcpy(dimensions[i].name, ConvertString2C(index_name));
				if(is_C)
				{
					dimensions[i].declared_start = va_arg(args, int);
					dimensions[i].declared_end = va_arg(args, int);
					dimensions[i].start = va_arg(args, int);
					dimensions[i].end = va_arg(args, int);
				}
				else
				{
					FORTRAN_INTEGER *value;

					value = va_arg(args, FORTRAN_INTEGER *);
					dimensions[i].declared_start = *value;
					value = va_arg(args, FORTRAN_INTEGER *);
					dimensions[i].declared_end = *value;
					value = va_arg(args, FORTRAN_INTEGER *);
					dimensions[i].start = *value;
					value = va_arg(args, FORTRAN_INTEGER *);
					dimensions[i].end = *value;
				}
				assert(dimensions[i].declared_start <= dimensions[i].declared_end);
				assert(dimensions[i].start <= dimensions[i].end);
				sprintf(buffer, "%d %s %d %d %d %d", i, dimensions[i].name,
					dimensions[i].declared_start, dimensions[i].declared_end,
					dimensions[i].start, dimensions[i].end);
#ifdef DEBUG
				fprintf(stderr, "ndregvar_: before reg dim %s\n", dimensions[i].name);
#endif
				TCPIPSendMessage(tcp_ip_handle, ND_REGISTER_DIM, buffer);
			} /* if(is_External) */
		}
	}
	else
		dimensions = NULL;
	
	va_end(args);

	if(is_External)
		TCPIPSendMessage(tcp_ip_handle, is_global ? ND_REGISTER_GLOBAL_VAR : ND_REGISTER_VAR, buffer);

	NewVarInfo(ConvertString2C(name), pointer, type, number_of_dimensions, dimensions, is_global);
}

/* Registration of variable "name".
   type - INTEGER_T, REAL_T, DOUBLE_T, or CHARACTER - (100 + len)
   number_of_dimensions - 0 if scalar.
   and then for each dimension:
   char *index_name;
   FORTRAN_INTEGER declared_start_point;
   FORTRAN_INTEGER declared_end_point;
   FORTRAN_INTEGER norma_start_point;
   FORTRAN_INTEGER norma_end_point;
   Norma                 Fortran              data
   B(i=1..40)            B(10)            "B\000", B, 1, 1, "i\000", 1, 40, 1, 10, 
   C(i=1..40; k=5..20)   C(0:11, 5:20)    "C\000", C, 1, 2, "i\000", 1, 40, 0, 11, "k\000", 5, 20, 5, 20
   N                     N                "N\000", N, 2, 0
*/

void NDRegisterVariable(const char *name, void *pointer, NormaVarType type, 
			  FORTRAN_INTEGER number_of_dimensions, ...)
{
	va_list args;
	va_start(args, number_of_dimensions);
	RegisterVariable(name, pointer, type, number_of_dimensions, TRUE, FALSE, FALSE, args);
}

void FNAME(ndregvar)(const char *name, void *pointer, NormaVarType *type, 
			  FORTRAN_INTEGER *number_of_dimensions, ...)
{
	va_list args;
	va_start(args, number_of_dimensions);
	RegisterVariable(name, pointer, *type, *number_of_dimensions, FALSE, FALSE, FALSE, args);
}

/* Registration of variable "name".
   type - INTEGER_T, REAL_T or DOUBLE_T
   number_of_dimensions - 0 if scalar.
   and then for each dimension:
   char *index_name;
   FORTRAN_INTEGER start_point;
   FORTRAN_INTEGER end_point;
*/
void FNAME(ndregvex)(const char *name, void *pointer, NormaVarType *type, 
			  FORTRAN_INTEGER *number_of_dimensions, ...)
{
	va_list args;
	va_start(args, number_of_dimensions);
	RegisterVariable(name, pointer, *type, *number_of_dimensions, FALSE, TRUE, FALSE, args);
}

/* Global Variable in External Fortran function
   The differnce with ndregvex() is only logical
 */
void FNAME(ndregvg)(const char *name, void *pointer, NormaVarType *type, 
			  FORTRAN_INTEGER *number_of_dimensions, ...)
{
	va_list args;
	va_start(args, number_of_dimensions);
	RegisterVariable(name, pointer, *type, *number_of_dimensions, FALSE, TRUE, TRUE, args);
}

/* called before each operator.
   operator == 0 => it is the first operator */
int NDOperatorStart(FORTRAN_INTEGER previous_operator,
			  FORTRAN_INTEGER previous_operator_line_begin,
			  FORTRAN_INTEGER previous_operator_line_end,
			  FORTRAN_INTEGER current_operator,
			  FORTRAN_INTEGER current_operator_line_begin,
			  FORTRAN_INTEGER current_operator_line_end)
{
	return FNAME(ndopstar)(&previous_operator,
			  &previous_operator_line_begin,
			  &previous_operator_line_end,
			  &current_operator,
			  &current_operator_line_begin,
			  &current_operator_line_end);
}


int FNAME(ndopstar)(FORTRAN_INTEGER *previous_operator,
			  FORTRAN_INTEGER *previous_operator_line_begin,
			  FORTRAN_INTEGER *previous_operator_line_end,
			  FORTRAN_INTEGER *current_operator,
			  FORTRAN_INTEGER *current_operator_line_begin,
			  FORTRAN_INTEGER *current_operator_line_end)
{
	char buffer[MAX_NUMBER_LEN * 6];

	assert(is_monitor_connect_inited);
	assert(cur_part_info);

	sprintf(buffer, FORTRAN_PRINT_STR6,
		*previous_operator,
		*previous_operator_line_begin,
		*previous_operator_line_end,
		*current_operator,
		*current_operator_line_begin,
		*current_operator_line_end);
	
#ifdef DEBUG
	fprintf(stderr, "ndopstar_: before operator %d\n", *current_operator);
#endif
	TCPIPSendMessage(tcp_ip_handle, ND_NOTIF_OPERATOR_START, buffer);
	GetMonitorAnswer();
	return 0;
}

int FNAME(ndopstex)(FORTRAN_INTEGER *is_potentially_lock,
			  FORTRAN_INTEGER *previous_operator_line_begin,
			  FORTRAN_INTEGER *previous_operator_line_end,
			  FORTRAN_INTEGER *current_operator_line_begin,
			  FORTRAN_INTEGER *current_operator_line_end)
{
	char buffer[MAX_NUMBER_LEN * 6];

	assert(is_monitor_connect_inited);
	assert(cur_part_info);

	sprintf(buffer, FORTRAN_PRINT_STR5,
		*is_potentially_lock,
		*previous_operator_line_begin,
		*previous_operator_line_end,
		*current_operator_line_begin,
		*current_operator_line_end);
	
#ifdef DEBUG
	fprintf(stderr, "ndopstex: before operator line %d\n", *current_operator_line_begin);
#endif
	TCPIPSendMessage(tcp_ip_handle, ND_NOTIF_EX_OPERATOR_START, buffer);
	GetMonitorAnswer();
	return 0;
}

/* called in start-part of Distribution part before send parameters and receive the rezult */
void NDWait(void) { FNAME(ndwait)();}
void FNAME(ndwait)(void)
{
	assert(is_monitor_connect_inited);
	TCPIPSendMessage(tcp_ip_handle, ND_WAIT_SIGNAL, NULL);
	GetMonitorAnswer();
}
/* called in Distribution part to notify that main part should continue */
void NDContinue(FORTRAN_INTEGER for_main) { FNAME(ndcont)(&for_main);}
void FNAME(ndcont)(FORTRAN_INTEGER *for_main)
{
	assert(is_monitor_connect_inited);
	if(*for_main)
		TCPIPSendMessage(tcp_ip_handle, ND_CONTINUE_SIGNAL1, NULL);
	else
		TCPIPSendMessage(tcp_ip_handle, ND_CONTINUE_SIGNAL0, NULL);
}
